Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
html_of_jsx
A JSX transformation and a library to write HTML in Reason and mlx.
Features
Brings the "component model" to HTML
Supports all of Reason's JSX features (uppercase components, fragments, optional attributes, punning) with a few improvements (lowercase components, no ppx annotation needed)
See Getting started for dune package manager, opam pin, and syntax setup.
Usage
let element = <a href="https://x.com/davesnx">
<span> {JSX.string("Click me!")} </span>
</a>;
let html = JSX.render(element);
/* <a href="https://x.com/davesnx"><span>Click me!</span></a> */
let element = <a href="https://x.com/davesnx">
<span>(JSX.string "Click me!")</span>
</a>
let html = JSX.render element
(* <a href="https://x.com/davesnx"><span>Click me!</span></a> *)
See Core API for elements, children, components, and rendering.
Credits
This library was extracted from server-reason-react and later simplified to work only with HTML5.